﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Xml.Serialization;
using System.IO;
using System.Diagnostics;
using Microsoft.Win32;
using System.ComponentModel;
using System.Runtime.InteropServices;
using System.Reflection;
using System.Drawing;
using System.Web;
using System.Text.RegularExpressions;
using System.Security.Principal;

namespace ROTools.ROWebLogin {

    /// <summary>
    /// メインフォーム
    /// </summary>
    public partial class MainForm : Form {

        /// <summary>
        /// 一発起動アカウント指定
        /// </summary>
        public string OneShotAccount {
            get;
            set;
        }

        /// <summary>
        /// 設定をまとめて保持するためのオブジェクト
        /// </summary>
        private AppConfigData appConfig = null;

        /// <summary>
        /// Web起動通信オブジェクト
        /// </summary>
        private WebGSComm webGSComm = null;

        /// <summary>
        /// 暗号キー
        /// </summary>
        private string cryptoPassPhrase = null;

        /// <summary>
        /// タスクトレイ用メニュー
        /// </summary>
        private ContextMenuStrip taskTrayMenu = null;

        /// <summary>
        /// タスクトレイ用メニュー
        /// </summary>
        private ToolStripMenuItem taskTrayLoginMenuItem = null;

        /// <summary>
        /// ログイン処理用ワーカーオブジェクト
        /// </summary>
        private BackgroundWorker loginWorker = null;

        /// <summary>
        /// パッチクライアント起動中フラグ
        /// </summary>
        private bool isPatchClientExecuting = false;

        /// <summary>
        /// ゲームクライアント起動中フラグ
        /// </summary>
        private bool isGameClientExecuting = false;

        /// <summary>
        /// 異常終了フラグ
        /// </summary>
        private bool isAbnormalTerminate = false;

        /// <summary>
        /// 電話認証コード
        /// </summary>
        private string ivrAuthCode = null;

        /// <summary>
        /// 電話認証中フォーム
        /// </summary>
        private IVRAuthCallingForm ivrAuthCalling = null;

        /// <summary>
        // EnumWindowsコールバックデリゲート
        /// </summary>
        /// <param name="hWnd"></param>
        /// <param name="lParam"></param>
        /// <returns></returns>
        private delegate int EnumerateWindowsCallback(IntPtr hWnd, int lParam);

        /// <summary>
        /// EnumWindows API
        /// </summary>
        /// <param name="lpEnumFunc"></param>
        /// <param name="lParam"></param>
        /// <returns></returns>
        [DllImport("user32", EntryPoint = "EnumWindows")]
        private static extern int EnumWindows(EnumerateWindowsCallback lpEnumFunc, int lParam);

        /// <summary>
        /// GetWindowText
        /// </summary>
        /// <param name="hWnd"></param>
        /// <param name="lpString"></param>
        /// <param name="nMaxCount"></param>
        /// <returns></returns>
        [DllImport("user32", EntryPoint = "GetWindowText", CharSet = CharSet.Auto)]
        private static extern int GetWindowText(IntPtr hWnd, StringBuilder lpString, int nMaxCount);

        /// <summary>
        /// IsWindowVisible
        /// </summary>
        /// <param name="hWnd"></param>
        /// <returns></returns>
        [DllImport("user32", EntryPoint = "IsWindowVisible")]
        private static extern int IsWindowVisible(IntPtr hWnd);

        private EnumerateWindowsCallback clientWindowEnumerator = null;

        /// <summary>
        /// ログイン処理用ワーカー引数クラス
        /// </summary>
        private class LoginWorkerArgs {
            /// <summary>
            /// パスワードの保存
            /// </summary>
            public bool SavePassword {
                get;
                set;
            }

            /// <summary>
            /// Urdrワールドに関する処理
            /// </summary>
            public bool IsUrdrWorld {
                get;
                set;
            }

            /// <summary>
            /// 処理対象アカウント
            /// </summary>
            public GunghoAccountData TargetGunghoAccount {
                get;
                set;
            }

            /// <summary>
            /// 処理対象アトラクション
            /// </summary>
            public AttractionAccountData TargetAttractionAccount {
                get;
                set;
            }
        }


        /// <summary>
        /// コンストラクタ
        /// </summary>
        public MainForm() {
            InitializeComponent();
            // Web起動用通信オブジェクトの生成
            webGSComm = new WebGSComm();
            webGSComm.OTPAuth += new EventHandler<OTPAuthEventArgs>(webGSComm_OTPAuth);
            webGSComm.IDLockAuth += new EventHandler<IDLockAuthEventArgs>(webGSComm_IDLockAuth);
            webGSComm.ImageAuth += new EventHandler<ImageAuthEventArgs>(webGSComm_ImageAuth);
            webGSComm.IVRAuthConfirm += new EventHandler<IVRAuthConfirmEventArgs>(webGSComm_IVRAuthConfirm);
            webGSComm.IVRAuthBeginCall += new EventHandler<EventArgs>(webGSComm_IVRAuthBeginCall);
            webGSComm.IVRAuthEndCall += new EventHandler<EventArgs>(webGSComm_IVRAuthEndCall);
            // ログイン処理用ワーカーオブジェクトの生成
            loginWorker = new BackgroundWorker();
            loginWorker.DoWork += new DoWorkEventHandler(loginWorker_DoWork);
            loginWorker.RunWorkerCompleted += new RunWorkerCompletedEventHandler(loginWorker_RunWorkerCompleted);
            // クライアントプロセス監視
            clientWindowEnumerator = new EnumerateWindowsCallback(ClientWindowEnumerator);
            return;
        }

        /// <summary>
        /// メッセージ表示
        /// </summary>
        /// <param name="message"></param>
        public void AddMessage(string message) {
            if(InvokeRequired) {
                object[] invokeArgs = new object[1];
                invokeArgs[0] = message;
                Invoke(new AddMessageDelegate(AddMessage), invokeArgs);
                return;
            }
            // TODO: あとでリストボックスにする？ 進捗状態を表示したい
            messageTextBox.Text += message;
            messageTextBox.SelectionStart = messageTextBox.Text.Length;
            messageTextBox.SelectionLength = 0;
            messageTextBox.ScrollToCaret();
        }

        public delegate void AddMessageDelegate(string message);


        /// <summary>
        /// 現在のアカウントデータを画面上に反映させる
        /// </summary>
        private void SetAccountDataToControls() {
            // ガンホーIDリストの更新
            gunghoIdComboBox.DataSource = null;
            gunghoIdComboBox.DataSource = appConfig.GunghoAccounts;
            gunghoIdComboBox.DisplayMember = "Id";
            gunghoIdComboBox.ValueMember = "Id";
            // メニューの更新
            taskTrayLoginMenuItem.DropDownItems.Clear();
            foreach(GunghoAccountData data in appConfig.GunghoAccounts) {
                if(data.AttractionAccounts != null) {
                    foreach(AttractionAccountData data2 in data.AttractionAccounts) {
                        ToolStripMenuItem item = new ToolStripMenuItem(data2.IdAndName);
                        item.Tag = data2;
                        taskTrayLoginMenuItem.DropDownItems.Add(item);
                    }
                }
            }
        }


        /// <summary>
        /// 管理者権限を持っているか判定する
        /// </summary>
        /// <returns></returns>
        private bool IsAdmin() {
            WindowsIdentity userId = WindowsIdentity.GetCurrent();
            WindowsPrincipal principal = new WindowsPrincipal(userId);
            return principal.IsInRole(@"BUILTIN\Administrators");
        }
        
        /// <summary>
        /// Vistaかどうか判定する
        /// </summary>
        /// <returns></returns>
        private bool IsVista() {
            OperatingSystem osInfo = Environment.OSVersion;
            if(osInfo.Version.Major >= 6) {
                return true;
            }
            return false;
        }


        /// <summary>
        /// クライアントを起動する
        /// </summary>
        /// <param name="command"></param>
        /// <param name="args"></param>
        /// <param name="account"></param>
        private void KickClient(string command, string args, AttractionAccountData account) {
            // TODO: ゲームを起動する処理は別クラスに分離するべきな気がしてきた。
            if(InvokeRequired) {
                object[] invokeArgs = new object[3];
                invokeArgs[0] = command;
                invokeArgs[1] = args;
                invokeArgs[2] = account;
                Invoke(new KickClientDelegate(KickClient), invokeArgs);
                return;
            }
            // Ragnarok.exeを起動する
            Debug.WriteLine("KickClient: start");
            AddMessage("KickClient: start\r\n");
            Process.Start("ROEXEURI://-w&" + account.Password);
        }

        private delegate void KickClientDelegate(string command, string args, AttractionAccountData account);


        private void KickClientWithLoadPrgAx(string gameCode, string coding, string attractionId, string password) {
            if(InvokeRequired) {
                object[] invokeArgs = new object[4];
                invokeArgs[0] = gameCode;
                invokeArgs[1] = coding;
                invokeArgs[2] = attractionId;
                invokeArgs[3] = password;
                Invoke(new KickClientWithLoadPrgAxDelegate(KickClientWithLoadPrgAx), invokeArgs);
                return;
            }
            Debug.WriteLine("KickClientWithLoadPrgAx: start");
            BrowserForm browserForm = new BrowserForm();
            browserForm.Show();
            WebBrowser browser = browserForm.webBrowser1;
            // 内容をクリアする
            browser.Navigate("about:blank", false);
            // ActiveXを起動するためのHTMLを書き込む
            StringBuilder html = new StringBuilder();
            html.Append("<html><body>\n");
            html.Append("<font color=\"blue\" size=\"2\">RagnarokOnline 起動中</font><br>\n");
            html.Append("<object id=\"LoadPrgAx\" codebase=\"https://member.gungho.jp/nhn/LoadPrgAx.CAB#version=1,0,0,18\" classid=\"CLSID:53F4962A-8E27-4601-8B01-79A82B4D7FC9\" width=\"1\" height=\"1\">\n");
            if(!string.IsNullOrEmpty(coding)) {
                html.Append("<param name=\"Coding\" value=\"HG\" />\n");
            }
            html.Append("<param name=\"GameCode\" value=\"");
            html.Append(gameCode);
            html.Append("\" />\n");
            html.Append("<param name=\"AttractionID\" value=\"");
            html.Append(attractionId);
            html.Append("\" />\n");
            html.Append("<param name=\"Password\" value=\"");
            html.Append(password);
            html.Append("\" />\n");
            html.Append("</object>\n");
            html.Append("</body></html>\n");
            browser.Document.Write(html.ToString());
        }

        private delegate void KickClientWithLoadPrgAxDelegate(string gameCode, string coding, string attractionId, string password);


        /// <summary>
        /// タイマーの動作状態を、設定内容によって変化させる
        /// </summary>
        private void SetTimerStateByConfig() {
            // もし、タイマーが不要ならば停止しておく
            if(string.IsNullOrEmpty(appConfig.OnPatchClientExecOperate) && string.IsNullOrEmpty(appConfig.OnGameClientTerminateOperate)) {
                clientCheckTimer.Stop();
            } else {
                clientCheckTimer.Start();
            }
        }


        /// <summary>
        /// ログイン処理用のワーカースレッドを起動する
        /// </summary>
        /// <param name="workerArgs"></param>
        /// <remarks>TargetGunghoAccountとTargetAttractionAccountを設定したうえで呼び出す</remarks>
        private void StartLoginWorker(LoginWorkerArgs workerArgs) {
            workerArgs.SavePassword = true;
            if(string.IsNullOrEmpty(workerArgs.TargetGunghoAccount.Password)) {
                using(AddIdForm form = new AddIdForm()) {
                    form.Location = appConfig.MainFormLocaiton;
                    form.Id = workerArgs.TargetGunghoAccount.Id;
                    form.Password = "";
                    form.IsOtpEnable = workerArgs.TargetGunghoAccount.IsOtpEnable;
                    if(form.ShowDialog() != DialogResult.OK) {
                        return;
                    }
                    workerArgs.TargetGunghoAccount.Password = form.Password;
                    workerArgs.TargetGunghoAccount.IsOtpEnable = form.IsOtpEnable;
                    if(!form.SavePassword) {
                        workerArgs.SavePassword = false;
                    }
                }
            }
            workerArgs.IsUrdrWorld = urdrCheckBox.Checked;
            // 各種ボタン類を操作できなくする
            gunghoIdComboBox.Enabled = false;
            attractionIdListBox.Enabled = false;
            addIdButton.Enabled = false;
            deleteIdButton.Enabled = false;
            configButton.Enabled = false;
            loginButton.Enabled = false;
            logoutButton.Enabled = false;
            urdrCheckBox.Enabled = false;
            // ログイン処理の開始
            loginWorker.RunWorkerAsync(workerArgs);
        }


        /// <summary>
        /// フォーム読み込み時処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainForm_Load(object sender, EventArgs e) {
            // バージョン表記
            FileVersionInfo versionInfo = FileVersionInfo.GetVersionInfo(Assembly.GetExecutingAssembly().Location);
            versionLabel.Text = "ver. " + versionInfo.FileVersion;
            // タスクトレイアイコンの準備
            taskTrayMenu = new ContextMenuStrip();
            taskTrayMenu.ItemClicked += new ToolStripItemClickedEventHandler(taskTrayMenu_ItemClicked);
            taskTrayIcon.ContextMenuStrip = taskTrayMenu;
            taskTrayLoginMenuItem = new ToolStripMenuItem("ログイン");
            taskTrayLoginMenuItem.DropDownItemClicked += new ToolStripItemClickedEventHandler(taskTrayLoginMenuItem_DropDownItemClicked);
            taskTrayMenu.Items.Add(taskTrayLoginMenuItem);
            taskTrayMenu.Items.Add("-");
            taskTrayMenu.Items.Add("元のサイズに戻す");
            taskTrayMenu.Items.Add("終了");
            // 設定の読み込み
            try {
                appConfig = AppConfigData.XmlFileDeserialize();
                if(appConfig.IsConfigCrypto) {
                    do {
                        using(PassPhraseInputForm form = new PassPhraseInputForm()) {
                            form.Location = appConfig.MainFormLocaiton;
                            if(form.ShowDialog() != DialogResult.OK) {
                                throw new PasswordRequireException();
                            }
                            cryptoPassPhrase = form.PassPhrase;
                        }
                        try {
                            appConfig = AppConfigData.CryptoFileDeserialize(cryptoPassPhrase);
                            Debug.WriteLine("passphrase OK");
                            break;
                        } catch {
                            Debug.WriteLine("passphrase NG");
                            // 例外が起きたらループ処理
                        }
                    } while(true);
                }
            } catch(PasswordRequireException) {
                isAbnormalTerminate = true;
                this.Close();
                return;
            } catch {
                appConfig = new AppConfigData();
                appConfig.GunghoAccounts = new List<GunghoAccountData>();
            }
            // フォームの状態を復元
            this.Location = appConfig.MainFormLocaiton;
            this.Size = appConfig.MainFormSize;
            this.urdrCheckBox.Checked = appConfig.IsUrdrWorld;
            SetAccountDataToControls();
            SetTimerStateByConfig();
            logoutButton.Enabled = false;
            // 一発起動処理
            GunghoAccountData oneShotGunghoAccount = null;
            AttractionAccountData oneShotAttractionAccount = null;
            if(!string.IsNullOrEmpty(OneShotAccount)) {
                foreach(GunghoAccountData data in appConfig.GunghoAccounts) {
                    if(data.AttractionAccounts != null) {
                        foreach(AttractionAccountData data2 in data.AttractionAccounts) {
                            if(data2.Id == OneShotAccount) {
                                oneShotGunghoAccount = data;
                                oneShotAttractionAccount = data2;
                                break;
                            }
                        }
                    }
                    if(oneShotGunghoAccount != null) {
                        break;
                    }
                }
                if(oneShotGunghoAccount == null) {
                    OneShotAccount = null;
                }
            }
            if(oneShotGunghoAccount != null) {
                gunghoIdComboBox.SelectedValue = oneShotGunghoAccount.Id;
            }
            gunghoIdComboBox_SelectedIndexChanged(null, null);
            if(oneShotAttractionAccount != null) {
                attractionIdListBox.SelectedValue = oneShotAttractionAccount.Id;
            }
            attractionIdListBox_SelectedIndexChanged(null, null);
            if(!string.IsNullOrEmpty(OneShotAccount)) {
                loginButton_Click(null, null);
            }
            return;
        }


        /// <summary>
        /// フォームを閉じるときの処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainForm_FormClosing(object sender, FormClosingEventArgs e) {
            // タイマーの停止
            clientCheckTimer.Stop();
            if(isAbnormalTerminate) {
                return;
            }
            // 設定を保存する
            appConfig.IsUrdrWorld = urdrCheckBox.Checked;
            try {
                if(appConfig.IsConfigCrypto) {
                    // 暗号化されていることを記録するだけのxmlファイルを出力
                    AppConfigData dummyConfig = new AppConfigData();
                    dummyConfig.IsConfigCrypto = true;
                    dummyConfig.MainFormLocaiton = appConfig.MainFormLocaiton;
                    dummyConfig.XmlFileSerialize();
                    // 実際の設定を出力
                    appConfig.CryptoFileSerialize(cryptoPassPhrase);
                } else {
                    appConfig.XmlFileSerialize();
                }
            } catch(Exception ex) {
                MessageBox.Show("設定の保存中にエラーが発生しました：" + ex.Message, this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            loginWorker.Dispose();
            return;
        }


        /// <summary>
        /// フォームのサイズ変更時処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainForm_Resize(object sender, EventArgs e) {
            if(this.WindowState == FormWindowState.Minimized) {
                this.Visible = false;
                taskTrayIcon.Visible = true;
            } else {
                if(!this.Visible) {
                    this.Visible = true;
                }
            }
        }


        /// <summary>
        /// フォームの移動時処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainForm_LocationChanged(object sender, EventArgs e) {
            if(this.Location.X != -32000 && this.Location.Y != -32000) {
                appConfig.MainFormLocaiton = this.Location;
            }
        }

        /// <summary>
        /// フォームのサイズ変更時処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void MainForm_SizeChanged(object sender, EventArgs e) {
            if(this.Size.Height >= this.MinimumSize.Height && this.Size.Width >= this.MinimumSize.Width) {
                appConfig.MainFormSize = this.Size;
            }
        }


        /// <summary>
        /// ログインボタンクリック時処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void loginButton_Click(object sender, EventArgs e) {
            if(loginWorker.IsBusy) {
                // ログイン処理の中断
                loginWorker.CancelAsync();
            } else {
                // ログイン処理に必要なデータの取得
                LoginWorkerArgs workerArgs = new LoginWorkerArgs();
                workerArgs.TargetGunghoAccount = (GunghoAccountData)gunghoIdComboBox.SelectedItem;
                if(workerArgs.TargetGunghoAccount == null) {
                    return;
                }
                workerArgs.TargetAttractionAccount = (AttractionAccountData)attractionIdListBox.SelectedItem;
                StartLoginWorker(workerArgs);
            }
        }


        /// <summary>
        /// ガンホーID追加ボタン
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void addIdButton_Click(object sender, EventArgs e) {
            using(AddIdForm form = new AddIdForm()) {
                form.Id = "";
                form.Password = "";
                form.Location = appConfig.MainFormLocaiton;
                if(form.ShowDialog() != DialogResult.OK) {
                    return;
                }
                GunghoAccountData data = new GunghoAccountData();
                data.Id = form.Id;
                if(form.SavePassword) {
                    data.Password = form.Password;
                } else {
                    data.Password = "";
                }
                data.IsOtpEnable = form.IsOtpEnable;
                appConfig.GunghoAccounts.Add(data);
                SetAccountDataToControls();
                gunghoIdComboBox.SelectedValue = data.Id;
            }
        }


        /// <summary>
        /// ガンホーID削除
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void deleteIdButton_Click(object sender, EventArgs e) {
            // 現在選択されているデータを取得
            GunghoAccountData targetGunghoAccountData = (GunghoAccountData)gunghoIdComboBox.SelectedItem;
            if(targetGunghoAccountData == null) {
                return;
            }
            if(MessageBox.Show("ガンホーID " + targetGunghoAccountData.Id + " を一覧から削除しますか？", this.Text, MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button2) != DialogResult.Yes) {
                return;
            }
            appConfig.GunghoAccounts.Remove(targetGunghoAccountData);
            SetAccountDataToControls();
            gunghoIdComboBox.SelectedIndex = gunghoIdComboBox.Items.Count - 1;
        }


        /// <summary>
        /// ガンホーID選択時の処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void gunghoIdComboBox_SelectedIndexChanged(object sender, EventArgs e) {
            attractionIdListBox.DataSource = null;
            attractionIdListBox.SelectedIndex = -1;
            if(gunghoIdComboBox.SelectedIndex < 0) {
                deleteIdButton.Enabled = false;
                return;
            }
            deleteIdButton.Enabled = true;
            GunghoAccountData targetData = (GunghoAccountData)gunghoIdComboBox.SelectedItem;
            attractionIdListBox.DataSource = targetData.AttractionAccounts;
            attractionIdListBox.DisplayMember = "IdAndName";
            attractionIdListBox.ValueMember = "Id";
            attractionIdListBox.SelectedIndex = -1;
        }


        /// <summary>
        /// アトラクションID選択時の処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void attractionIdListBox_SelectedIndexChanged(object sender, EventArgs e) {
            expiryDateLabel.Text = "";
            if(attractionIdListBox.SelectedIndex < 0) {
                loginButton.Text = "アトラクションID取得";
            } else {
                loginButton.Text = "ログイン";
                AttractionAccountData data = (AttractionAccountData)attractionIdListBox.SelectedItem;
                string expiryDate = data.ExpiryDateString;
                if(!string.IsNullOrEmpty(expiryDate)) {
                    Regex regex = new Regex("^利用期限 ： (.+)\\(.+\\)");
                    Match match = regex.Match(expiryDate);
                    if(match.Success) {
                        expiryDateLabel.Text = match.Groups[1].Value;
                    } else {
                        expiryDateLabel.Text = expiryDate;
                    }
                }
            }
        }


        /// <summary>
        /// アトラクションIDをダブルクリックしたとき
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void attractionIdListBox_DoubleClick(object sender, EventArgs e) {
            if(attractionIdListBox.SelectedIndex < 0 || loginButton.Enabled == false) {
                return;
            }
            loginButton_Click(sender, e);
        }



        /// <summary>
        /// 設定ボタンをクリックしたとき
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void configButton_Click(object sender, EventArgs e) {
            using(ConfigForm form = new ConfigForm()) {
                form.CryptoPassPhrase = cryptoPassPhrase;
                form.AppConfig = appConfig;
                form.ShowDialog();
                cryptoPassPhrase = form.CryptoPassPhrase;
            }
            SetTimerStateByConfig();
        }


        /// <summary>
        /// ログイン処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void loginWorker_DoWork(object sender, DoWorkEventArgs e) {
            Debug.WriteLine("loginWorker_DoWork: start");
            LoginWorkerArgs workerArgs = (LoginWorkerArgs)e.Argument;
            e.Result = workerArgs;
            try {
                // 選択されているアトラクションIDでログインし、アトラクションID一覧を取得する
                webGSComm.AccountData = workerArgs.TargetGunghoAccount;
                webGSComm.GetAttractionAccountData();
                if(workerArgs.TargetAttractionAccount == null) {
                    // ここで処理を終了
                    return;
                }
                // 指定されたアトラクションIDが存在しているか確認
                bool attractionIdExists = false;
                foreach(AttractionAccountData checkData in workerArgs.TargetGunghoAccount.AttractionAccounts) {
                    if(checkData.InternalId == workerArgs.TargetAttractionAccount.InternalId) {
                        workerArgs.TargetAttractionAccount = checkData; //念のため
                        attractionIdExists = true;
                        break;
                    }
                }
                if(!attractionIdExists) {
                    // アトラクションIDが消えていた
                    throw new AppException("指定されたアトラクションIDは存在しませんでした");
                }
                // パスワードの取得
                webGSComm.GetOneTimePassword(workerArgs.TargetAttractionAccount.InternalId, workerArgs.IsUrdrWorld);
                Debug.WriteLine("loginWorker_DoWork: success");
            } catch(Exception ex) {
                e.Result = ex;
            } finally {
                if(!workerArgs.SavePassword) {
                    // パスワードを消しておく
                    workerArgs.TargetGunghoAccount.Password = "";
                }
            }
        }


        /// <summary>
        /// ログイン処理完了時
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void loginWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e) {
            try {
                Debug.WriteLine("loginWorker_RunWorkerCompleted: start");
                // 各種ボタン類を操作できるようにする
                gunghoIdComboBox.Enabled = true;
                attractionIdListBox.Enabled = true;
                attractionIdListBox.SelectedItem = null;
                addIdButton.Enabled = true;
                deleteIdButton.Enabled = true;
                configButton.Enabled = true;
                loginButton.Enabled = true;
                urdrCheckBox.Enabled = true;
                // 処理結果を確認
                if(e.Result is Exception) {
                    Exception ex = (Exception)e.Result;
                    StringBuilder msg = new StringBuilder();
                    msg.Append("例外発生1: ");
                    msg.Append(ex.Message);
                    msg.Append("\r\n");
                    AddMessage(msg.ToString());
                    AddMessage(" *Source=" + ex.Source + "\r\n");
                    AddMessage(" *StackTrace=" + ex.StackTrace + "\r\n");
                    Debug.WriteLine(msg.ToString());
                    Debug.WriteLine(ex.StackTrace);
                    MessageBox.Show(msg.ToString(), this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }
                LoginWorkerArgs workerArgs = null;
                if(e.Result is LoginWorkerArgs) {
                    workerArgs = (LoginWorkerArgs)e.Result;
                }
                if(workerArgs == null) {
                    Debug.WriteLine("loginWorker_RunWorkerCompleted: no result");
                    return;
                }
                // アカウント情報を更新
                SetAccountDataToControls();
                gunghoIdComboBox.SelectedValue = workerArgs.TargetGunghoAccount.Id;
                if(workerArgs.TargetAttractionAccount != null) {
                    attractionIdListBox.SelectedValue = workerArgs.TargetAttractionAccount.Id;
                } else {
                    // 起動対象のアトラクションIDが指定されていないときは、ここで終了する
                    // (選択しやすいように、アトラクションIDリストにフォーカスを移動する)
                    attractionIdListBox.Focus();
                    if(attractionIdListBox.Items.Count > 0) {
                        attractionIdListBox.SelectedIndex = 0; // 必要かなぁ？人それぞれな気がする。
                    }
                    // 自動ログアウト
                    if(appConfig.IsAutoLogoutExec) {
                        webGSComm.Logout();
                    }
                    return;
                }
                // 有効期限切れ警告
                if(workerArgs.TargetAttractionAccount.Expired) {
                    if(MessageBox.Show("アトラクションIDの有効期限が切れています。\r\nクライアントを起動しますか？", this.Text, MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button2) == DialogResult.No) {
                        return;
                    }
                }
                // ゲームを起動
                string cmd = null;
                if(workerArgs.IsUrdrWorld) {
                    cmd = "RagnarokUrdr.exe";
                } else {
                    cmd = "Ragnarok.exe";
                }
                StringBuilder cmdArg = new StringBuilder();
                cmdArg.Append("-w -u:");
                cmdArg.Append(workerArgs.TargetAttractionAccount.Id);
                cmdArg.Append(" -p:");
                cmdArg.Append(workerArgs.TargetAttractionAccount.Password);
                KickClient(cmd, cmdArg.ToString(), workerArgs.TargetAttractionAccount);
                // 自動ログアウト
                if(appConfig.IsAutoLogoutExec) {
                    webGSComm.Logout();
                }
                // 一発起動処理の場合は、ここでアプリケーションを終了する
                if(!string.IsNullOrEmpty(OneShotAccount)) {
                    this.Close();
                }
            } catch(Exception ex) {
                StringBuilder msg = new StringBuilder();
                msg.Append("例外発生2: ");
                msg.Append(ex.Message);
                msg.Append("\r\n");
                AddMessage(msg.ToString());
                AddMessage(" *Source=" + ex.Source + "\r\n");
                AddMessage(" *StackTrace=" + ex.StackTrace + "\r\n");
                Debug.WriteLine(msg.ToString());
                Debug.WriteLine(ex.StackTrace);
                MessageBox.Show(msg.ToString(), this.Text, MessageBoxButtons.OK, MessageBoxIcon.Error);
            } finally {
                OneShotAccount = null;
                if(webGSComm.IsLoggedIn) {
                    logoutButton.Enabled = true;
                }
            }
        }


        /// <summary>
        /// 
        /// </summary>
        /// <param name="hWnd"></param>
        /// <param name="lParam"></param>
        /// <returns></returns>
        private int ClientWindowEnumerator(IntPtr hWnd, int lParam) {
            StringBuilder sb = new StringBuilder(0xfff);
            if(IsWindowVisible(hWnd) != 0) {
                // キャプションを取得
                if(GetWindowText(hWnd, sb, 0xfff) != 0) {
                    // Ragnarokクライアントかどうか
                    if(sb.ToString() == "Ragnarok") {
                        isGameClientExecuting = true;
                        return 0; // ここで終了
                    }
                }
            }
            return 1; // 継続
        }


        /// <summary>
        /// クライアント起動確認タイマー処理
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void clientCheckTimer_Tick(object sender, EventArgs e) {
            // パッチクライアントが起動しているか確認する
            int procNum = Process.GetProcessesByName("RAGNAROK").Length;
            if(procNum == 0) {
                procNum = Process.GetProcessesByName("RAGNAROKURDR").Length;
            }
            if(isPatchClientExecuting) {
                if(procNum == 0) {
                    // パッチクライアントが終了したときの処理
                    isPatchClientExecuting = false;
                }
            } else {
                if(procNum > 0) {
                    // パッチクライアントが起動したときの処理
                    isPatchClientExecuting = true;
                    switch(appConfig.OnPatchClientExecOperate) {
                        case "Terminate":
                            this.Close();
                            break;
                        case "Minimize":
                            this.WindowState = FormWindowState.Minimized;
                            break;
                    }
                }
            }


            // ゲームクライアントが起動しているか確認する
#if true
            bool currentlyGameClientExecuting = isGameClientExecuting;
            isGameClientExecuting = false;
            EnumWindows(clientWindowEnumerator, 0);
            if(currentlyGameClientExecuting) {
                if(!isGameClientExecuting) {
                    // ゲームクライアントが終了したときの処理
                    switch(appConfig.OnGameClientTerminateOperate) {
                        case "Normalize":
                            this.WindowState = FormWindowState.Normal;
                            taskTrayIcon_DoubleClick(null, null);
                            break;
                    }
                }
            } else {
                if(isGameClientExecuting) {
                    // ゲームクライアントが起動したときの処理
                }
            }
#else
            procNum = Process.GetProcessesByName("RAGEXE").Length;
            if(procNum == 0) {
                procNum = Process.GetProcessesByName("RAGURDREXE").Length;
            }
            if(isGameClientExecuting) {
                if(procNum == 0) {
                    // ゲームクライアントが終了したときの処理
                    isGameClientExecuting = false;
                    switch(appConfig.OnGameClientTerminateOperate) {
                        case "Normalize":
                            this.WindowState = FormWindowState.Normal;
                            taskTrayIcon_DoubleClick(null, null);
                            break;
                    }
                }
            } else {
                if(procNum > 0) {
                    // ゲームクライアントが起動したときの処理
                    isGameClientExecuting = true;
                }
            }
#endif

        }


        /// <summary>
        /// タスクトレイアイコンのダブルクリック時
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void taskTrayIcon_DoubleClick(object sender, EventArgs e) {
            this.Visible = true;
            this.WindowState = FormWindowState.Normal;
            this.Location = appConfig.MainFormLocaiton;
            taskTrayIcon.Visible = false;
        }


        /// <summary>
        /// タスクトレイメニューの項目選択時
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void taskTrayMenu_ItemClicked(object sender, ToolStripItemClickedEventArgs e) {
            ToolStripItem item = e.ClickedItem;
            if(item.Text == "元のサイズに戻す") {
                taskTrayIcon_DoubleClick(null, null);
            } else if(item.Text == "終了") {
                this.Close();
            }
        }


        /// <summary>
        /// タスクトレイメニューの項目選択時
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void taskTrayLoginMenuItem_DropDownItemClicked(object sender, ToolStripItemClickedEventArgs e) {
            ToolStripMenuItem item = e.ClickedItem as ToolStripMenuItem;
            if(item == null) {
                return;
            }
            if(item.Tag == null) {
                return;
            }
            // アカウントデータから、起動処理に必要となるデータを集める
            LoginWorkerArgs workerArgs = new LoginWorkerArgs();
            workerArgs.TargetAttractionAccount = (AttractionAccountData)item.Tag;
            workerArgs.TargetGunghoAccount = null;
            foreach(GunghoAccountData data in appConfig.GunghoAccounts) {
                if(data.AttractionAccounts != null) {
                    if(data.AttractionAccounts.Contains(workerArgs.TargetAttractionAccount)) {
                        workerArgs.TargetGunghoAccount = data;
                        break;
                    }
                }
            }
            if(workerArgs.TargetGunghoAccount == null) {
                return;
            }
            // ログイン処理を読み出す
            StartLoginWorker(workerArgs);
        }

        /// <summary>
        /// ログアウトボタン
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void logoutButton_Click(object sender, EventArgs e) {
            webGSComm.Logout();
            logoutButton.Enabled = webGSComm.IsLoggedIn;
        }

        private delegate void OTPAuthDelegate(object sender, OTPAuthEventArgs e);

        /// <summary>
        /// OTP認証
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void webGSComm_OTPAuth(object sender, OTPAuthEventArgs e) {
            if(this.InvokeRequired) {
                this.Invoke(new OTPAuthDelegate(webGSComm_OTPAuth), new object[] { sender, e });
            } else {
                using(OtpInputForm form = new OtpInputForm()) {
                    form.TopMost = true;
                    if(form.ShowDialog() == DialogResult.OK) {
                        e.OTP = form.Otp;
                    }
                }
            }
        }

        private delegate void ImageAuthDelegate(object sender, ImageAuthEventArgs e);

        /// <summary>
        /// 画像認証
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void webGSComm_ImageAuth(object sender, ImageAuthEventArgs e) {
            if(this.InvokeRequired) {
                this.Invoke(new ImageAuthDelegate(webGSComm_ImageAuth), new object[] { sender, e });
            } else {
                using(ImageAuthInputForm form = new ImageAuthInputForm()) {
                    form.CaptchaImage = e.CaptchaImage;
                    form.SaveCookie = e.SaveAuthState;
                    if(form.ShowDialog() == DialogResult.OK) {
                        e.CaptchaText = form.CaptchaText;
                        e.SaveAuthState = form.SaveCookie;
                    }
                }
            }
        }

        private delegate void IDLockAuthDelegate(object sender, IDLockAuthEventArgs e);

        /// <summary>
        /// IDロック認証
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void webGSComm_IDLockAuth(object sender, IDLockAuthEventArgs e) {
            if(this.InvokeRequired) {
                this.Invoke(new IDLockAuthDelegate(webGSComm_IDLockAuth), new object[] { sender, e });
            } else {
                using(IDLockAuthCodeInputForm form = new IDLockAuthCodeInputForm()) {
                    form.TopMost = true;
                    if(form.ShowDialog() == DialogResult.OK) {
                        e.AuthCode = form.AuthCode;
                    }
                }
            }
        }

        private delegate void IVRAuthConfirmDelegate(object sender, IVRAuthConfirmEventArgs e);
        private delegate void IVRAuthCallDelegate(object sender, EventArgs e);

        private void webGSComm_IVRAuthConfirm(object sender, IVRAuthConfirmEventArgs e) {
            if(this.InvokeRequired) {
                this.Invoke(new IVRAuthConfirmDelegate(webGSComm_IVRAuthConfirm), new object[] { sender, e });
            } else {
                Debug.WriteLine("webGSComm_IVRAuthConfirm: code=" + e.AuthCode);
                ivrAuthCode = e.AuthCode;
                using(IVRAuthConfirmForm form = new IVRAuthConfirmForm()) {
                    form.AuthCode = e.AuthCode;
                    if(form.ShowDialog() != DialogResult.OK) {
                        e.Cancel = true;
                    }
                }
            }
        }

        private void webGSComm_IVRAuthBeginCall(object sender, EventArgs e) {
            if(this.InvokeRequired) {
                this.Invoke(new IVRAuthCallDelegate(webGSComm_IVRAuthBeginCall), new object[] { sender, e });
            } else {
                Debug.WriteLine("webGSComm_IVRAuthBeginCall");
                if(ivrAuthCalling != null) {
                    if(!ivrAuthCalling.IsDisposed) {
                        ivrAuthCalling.Dispose();
                    }
                }
                ivrAuthCalling = new IVRAuthCallingForm();
                ivrAuthCalling.Owner = this;
                ivrAuthCalling.AuthCode = (ivrAuthCode == null) ? "(不明)" : ivrAuthCode;
                ivrAuthCalling.Show();
            }
        }

        private void webGSComm_IVRAuthEndCall(object sender, EventArgs e) {
            if(this.InvokeRequired) {
                this.Invoke(new IVRAuthCallDelegate(webGSComm_IVRAuthEndCall), new object[] { sender, e });
            } else {
                Debug.WriteLine("webGSComm_IVRAuthEndCall");
                ivrAuthCode = null;
                ivrAuthCalling.Close();
                ivrAuthCalling.Dispose();
                ivrAuthCalling = null;
            }
        }

        /// <summary>
        /// デバッグ用
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void versionLabel_MouseClick(object sender, MouseEventArgs e) {
#if DEBUG
#endif
        }

    }
}
