﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.IO;
using System.Diagnostics;
using System.Web;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using System.Collections.Specialized;
using System.Drawing;

namespace ROTools.ROWebLogin {

    /// <summary>
    /// GunghoのWebサーバと通信して、ゲーム起動用の処理を行う
    /// </summary>
    /// 
    public class WebGSComm {

        private const string URL_GUNGHO = @"https://member.gungho.jp";
        private const string URL_FRONT_LOGIN_PATH = @"/front/ro/iframe/login.aspx";
        private const string URL_FRONT_LOGIN = URL_GUNGHO + URL_FRONT_LOGIN_PATH;
        private const string URL_FRONT_INFO_PATH = @"/front/info.aspx";
        private const string URL_FRONT_WEBGS_PATH = @"/front/ro/iframe/menu.aspx";
        private const string URL_FRONT_WEBGS = URL_GUNGHO + URL_FRONT_WEBGS_PATH;
        private const string URL_FRONT_WEBGS_HG = URL_GUNGHO + @"/nhn/ragnarok/center.aspx";
        private const string URL_FRONT_SL_AUTH = URL_GUNGHO + @"/front/safetylock/authentication.aspx";
        private const string URL_FRONT_SL_VERIFY = URL_GUNGHO + @"/front/safetylock/verify.aspx";
        private const string URL_FRONT_SL_PRENOTEIVR_PATH = @"/front/safetylock/prenoteivr.aspx";
        private const string URL_FRONT_SL_PRENOTEIVR = URL_GUNGHO + URL_FRONT_SL_PRENOTEIVR_PATH;
        private const string URL_FRONT_SL_IVRAUTH_PATH = @"/front/safetylock/ivrauth.aspx";
        private const string URL_FRONT_SL_IVRAUTH = URL_GUNGHO + URL_FRONT_SL_IVRAUTH_PATH;
        private const string URL_FRONT_SL_FINISH = URL_GUNGHO + @"/front/safetylock/finish.aspx";
        private const string URL_FRONT_IA_AUTH = URL_GUNGHO + @"/front/ro/guest/imageauth.aspx";
        private const string URL_FRONT_IA_REDIRECT_PATH = @"/front/ro/iframe/redirectimageauth.aspx";
        private const string URL_FRONT_IA_IMAGE = URL_GUNGHO + @"/front/register/JpegImage.aspx";
        private const string URL_FRONT_IL_REDIRECT_PATH = @"/front/ro/iframe/redirectidlock.aspx";
        private const string URL_FRONT_IL_REDIRECT = URL_GUNGHO + URL_FRONT_IL_REDIRECT_PATH;
        private const string URL_FRONT_LAUNCH = URL_GUNGHO + @"/front/ro/member/GameLounch.aspx";
        private const string URL_FRONT_LAUNCH_REDIRECT_PATH = @"/front/ro/iframe/redirectOnGameLaunch.aspx";
        private const string URL_FRONT_TERM = URL_GUNGHO + @"/front/ro/member/TermInformation.aspx";
        private const string URL_HG_RAGNAROK = @"http://ragnarokonline.hangame.co.jp/index.nhn";
        private const string URL_HG_LOGIN = @"https://id.hangame.co.jp/login.asp";
        private const string URL_HG_WEBGS = @"http://ragnarokonline.hangame.co.jp/externalStartGame.nhn";


        // TODO: パスワード取得処理部分の見直し。(ログイン前提になっている)

        /// <summary>
        /// クッキー保持用オブジェクト
        /// </summary>
        private CookieContainer cookieContainer;

        private GunghoAccountData _accountData;

        /// <summary>
        /// 処理対象となるアカウント情報
        /// </summary>
        public GunghoAccountData AccountData {
            get {
                return _accountData;
            }
            set {
                if(_accountData != value) {
                    cookieContainer = new CookieContainer();
                }
                _accountData = value;
            }
        }

        /// <summary>
        /// 最後に取得できたViewState
        /// </summary>
        private string webGSViewState;

        /// <summary>
        /// 現在の状態
        /// </summary>
        public string CurrentStateMessage {
            get;
            set;
        }

        /// <summary>
        /// 現在アトラクションセンターへログイン中かどうか取得する
        /// </summary>
        /// <remarks>
        /// 付け足しプロパティなため、正確な状態は取得できない。これがtrueになるのは、
        /// 「アトラクションセンターからログアウトするために必要な情報が揃っている」
        /// 状態、つまり「ログアウト可能」な状態である。
        /// </remarks>
        public bool IsLoggedIn {
            get {
                if(cookieContainer == null) {
                    return false;
                }
                if(string.IsNullOrEmpty(webGSViewState)) {
                    return false;
                }
                return true;
            }
        }

        /// <summary>
        /// ワンタイムパスワード入力要求イベント
        /// </summary>
        public event EventHandler<OTPAuthEventArgs> OTPAuth;

        /// <summary>
        /// IDロック認証コード入力要求イベント
        /// </summary>
        public event EventHandler<IDLockAuthEventArgs> IDLockAuth;

        /// <summary>
        /// 画像認証コード入力要求イベント
        /// </summary>
        public event EventHandler<ImageAuthEventArgs> ImageAuth;

        /// <summary>
        /// 電話認証実行確認要求イベント
        /// </summary>
        public event EventHandler<IVRAuthConfirmEventArgs> IVRAuthConfirm;

        /// <summary>
        /// 電話認証発信イベント
        /// </summary>
        public event EventHandler<EventArgs> IVRAuthBeginCall;

        /// <summary>
        /// 電話認証終話イベント
        /// </summary>
        public event EventHandler<EventArgs> IVRAuthEndCall;

        /// <summary>
        /// コンストラクタ
        /// </summary>
        public WebGSComm() {
            CurrentStateMessage = null;
            ServicePointManager.SecurityProtocol = SecurityProtocolTypeExtensions.Tls12;
        }

        /// <summary>
        /// アトラクションセンターのログイン処理を行う
        /// </summary>
        public void Login() {
            Login(false);
        }

        /// <summary>
        /// アトラクションセンターのログイン処理を行う
        /// </summary>
        /// <param name="recursive"></param>
        public void Login(bool recursive) {
            CurrentStateMessage = "アトラクションセンターへのログインを開始";
            if(AccountData == null) {
                throw new AppException("ログイン用のデータが設定されていません");
            }
            if(string.IsNullOrEmpty(AccountData.Id) || string.IsNullOrEmpty(AccountData.Password)) {
                throw new AppException("ログイン用のID/パスワードが設定されていません");
            }
            // ハンゲーム用IDだったら、そちらへ処理をまわす
            if(AccountData.Id.StartsWith("HG:")) {
                LoginHanGame();
                return;
            }
            // とりあえず、ASP.NETのセッションキーとVIEWSTATEをもらってくる
            CurrentStateMessage = "アトラクションセンターからCookieを取得中";
            if(!recursive) {
                cookieContainer = new CookieContainer(); // Cookieを初期化
                if(AccountData.IDLockCookie != null) {
#if DEBUG
                    StringBuilder sb = new StringBuilder();
                    sb.Append("Login: add IDLockCookie:\r\n");
                    sb.Append(" Name=" + AccountData.IDLockCookie.Name + "\r\n");
                    sb.Append(" Value=" + AccountData.IDLockCookie.Value + "\r\n");
                    sb.Append(" Domain=" + AccountData.IDLockCookie.Domain + "\r\n");
                    sb.Append(" Port=" + AccountData.IDLockCookie.Port + "\r\n");
                    sb.Append(" Path=" + AccountData.IDLockCookie.Path + "\r\n");
                    sb.Append(" Secure=" + AccountData.IDLockCookie.Secure + "\r\n");
                    sb.Append(" Expires=" + AccountData.IDLockCookie.Expires + "\r\n");
                    sb.Append(" Expired=" + AccountData.IDLockCookie.Expired);
                    Debug.WriteLine(sb.ToString());
#endif
                    if(AccountData.IDLockCookie.Expired) {
                        Debug.WriteLine(" *** IDLockCookie=Expired");
                        AccountData.IDLockCookie = null;
                    } else {
                        cookieContainer.Add(AccountData.IDLockCookie);
                    }
                }
                if(AccountData.ImageAuthCookie != null) {
#if DEBUG
                    StringBuilder sb = new StringBuilder();
                    sb.Append("Login: add ImageAuthCookie:\r\n");
                    sb.Append(" Name=" + AccountData.ImageAuthCookie.Name + "\r\n");
                    sb.Append(" Value=" + AccountData.ImageAuthCookie.Value + "\r\n");
                    sb.Append(" Domain=" + AccountData.ImageAuthCookie.Domain + "\r\n");
                    sb.Append(" Port=" + AccountData.ImageAuthCookie.Port + "\r\n");
                    sb.Append(" Path=" + AccountData.ImageAuthCookie.Path + "\r\n");
                    sb.Append(" Secure=" + AccountData.ImageAuthCookie.Secure + "\r\n");
                    sb.Append(" Expires=" + AccountData.ImageAuthCookie.Expires + "\r\n");
                    sb.Append(" Expired=" + AccountData.ImageAuthCookie.Expired);
                    Debug.WriteLine(sb.ToString());
#endif
                    if(AccountData.ImageAuthCookie.Expired) {
                        Debug.WriteLine(" *** ImageAuthCookie=Expired");
                        AccountData.ImageAuthCookie = null;
                    } else if(AccountData.ImageAuthCookie.Expires.Equals(new DateTime(1, 1, 1))) {
                        // 1.2.4.0で誤ったCookie情報を保持したままになっている状況を解消するため
                        Debug.WriteLine(" *** ImageAuthCookie=Invalid");
                        AccountData.ImageAuthCookie = null;
                    } else {
                        cookieContainer.Add(AccountData.ImageAuthCookie);
                        cookieContainer.Add(new Cookie("GHLS", "0", AccountData.ImageAuthCookie.Path, AccountData.ImageAuthCookie.Domain));
                    }
                }
            }
            webGSViewState = null;
            HttpWebRequest request = CreateHttpWebRequest(URL_FRONT_LOGIN);
#if DEBUG
            foreach(Cookie dck in request.CookieContainer.GetCookies(request.RequestUri)) {
                Debug.WriteLine("Login: send cookie: " + dck.Name + "=" + dck.Value);
            }
#endif
            HttpWebResponse response = (HttpWebResponse)request.GetResponse();
            Debug.WriteLine("Login: response1=" + response.StatusCode);
#if DEBUG
            foreach(Cookie dck in response.Cookies) {
                Debug.WriteLine("Login: recv cookie: " + dck.Name + "=" + dck.Value);
            }
#endif
            if(response.StatusCode != HttpStatusCode.OK) {
                response.Close();
                throw new AppException("アトラクションセンターへのログイン中にエラーが発生しました");
            }
            string viewStateString = null;
            using(StreamReader reader = new StreamReader(response.GetResponseStream(), Encoding.GetEncoding(response.CharacterSet))) {
                Regex regex = new Regex("^.*<input type=\"hidden\" name=\"__VIEWSTATE\" id=\"__VIEWSTATE\" value=\"(.+)\" />.*$");
                string line = null;
                Match match = null;
                while((line = reader.ReadLine()) != null) {
                    match = regex.Match(line);
                    if(match.Success) {
                        viewStateString = match.Groups[1].Value;
                        break;
                    }
                }
            }
            response.Close();
            if(viewStateString == null) {
                throw new AppException("アトラクションセンターへのログインで、VIEWSTATEが取得できませんでした");
            }
            // OTPを利用する場合は、ここで入力を受け付ける
            string otp = null;
            if(AccountData.IsOtpEnable && OTPAuth != null) {
                OTPAuthEventArgs e = new OTPAuthEventArgs();
                OTPAuth(this, e);
                otp = e.OTP;
            }
            // ログインに必要な情報を投げつける
            CurrentStateMessage = "アトラクションセンターへ " + AccountData.Id + " でログイン中";
            request = CreateHttpWebRequest(URL_FRONT_LOGIN);
            request.Method = "POST";
            request.ContentType = "application/x-www-form-urlencoded";
            request.Referer = URL_FRONT_LOGIN;
            StringBuilder postData = new StringBuilder();
            postData.Append("__EVENTTARGET=");
            postData.Append("&__EVENTARGUMENT=");
            postData.Append("&__VIEWSTATE=");
            postData.Append(HttpUtility.UrlEncode(viewStateString));
            postData.Append("&loginNameControl%24txtLoginName=");
            postData.Append(HttpUtility.UrlEncode(AccountData.Id.Substring(0, (AccountData.Id.Length < 32) ? AccountData.Id.Length : 32)));
            postData.Append("&login=");
            postData.Append("&passwordControl%24txtPassword=");
            postData.Append(HttpUtility.UrlEncode(AccountData.Password.Substring(0, (AccountData.Password.Length < 16) ? AccountData.Password.Length : 16)));
            postData.Append("&OTPControl%24inputOTP=");
            if(!string.IsNullOrEmpty(otp)) {
                postData.Append(HttpUtility.UrlEncode(otp));
            }
            byte[] postBytes = Encoding.ASCII.GetBytes(postData.ToString());
            request.ContentLength = postBytes.Length;
            using(Stream postStream = request.GetRequestStream()) {
                postStream.Write(postBytes, 0, postBytes.Length);
            }
            response = (HttpWebResponse)request.GetResponse();
            Debug.WriteLine("Login: response2=" + response.StatusCode);
            string resultCode = null;
            string resultMessage = null;
            if(response.StatusCode == HttpStatusCode.Found) {
                // どこかに飛ばされた場合は、ログイン成功？
                if(response.Headers["Location"] == URL_FRONT_WEBGS_PATH) {
                    CurrentStateMessage = "アトラクションセンターへのログインが完了";
                    Debug.WriteLine("Login: login successful");
                    resultCode = "success";
                } else if(response.Headers["Location"].StartsWith(URL_FRONT_IL_REDIRECT_PATH)) {
                    if(response.Headers["Location"].Contains("&otp=")) {
                        Debug.WriteLine("Login: ID-Lock(IVR)");
                        resultCode = "idLock2";
                        resultMessage = response.Headers["Location"]; // とりあえずAuthCode取得用にデータを取っておく.
                    } else {
                        // 電話番号登録を促す画面への遷移とみなす
                        CurrentStateMessage = "アトラクションセンターへのログインが完了(IDLOCK?)";
                        Debug.WriteLine("Login: login successful (IDLOCK?)");
                        resultCode = "success";
                    }
                } else if(response.Headers["Location"].StartsWith(URL_FRONT_IA_REDIRECT_PATH)) {
                    Debug.WriteLine("Login: require imageauth");
                    resultCode = "imageauth";
                    resultMessage = response.Headers["Location"]; // とりあえずotp取得用にデータを取っておく.
                } else {
                    resultMessage = "アトラクションセンターへのログイン中に不明なリダイレクトが返されました (" + response.Headers["Location"] + ")";
                }
            } else if(response.StatusCode == HttpStatusCode.OK) {
                // 200応答の場合は、ログインに失敗しているので原因を取得する
                Debug.WriteLine("Login: Response=200");
                using(StreamReader reader = new StreamReader(response.GetResponseStream(), Encoding.GetEncoding(response.CharacterSet))) {
                    Regex idLockRegex = new Regex("^.*<span id=\"labInfomation\" class=\"idlockmsg\">.*IDロックが設定されています。.*$");
                    Regex errorRegex = new Regex("^.*<p id=\"err\">.*$");
                    string line = null;
                    Match match = null;
                    while((line = reader.ReadLine()) != null) {
                        // IDロックメッセージの確認
                        match = idLockRegex.Match(line);
                        if(match.Success) {
                            resultCode = "idLock";
                            break;
                        }
                        // エラーメッセージの確認
                        match = errorRegex.Match(line);
                        if(match.Success) {
                            // 次の行がエラーメッセージ
                            string errorMessage = reader.ReadLine();
                            if(!string.IsNullOrEmpty(errorMessage)) {
                                errorMessage = errorMessage.Trim();
                                if(!string.IsNullOrEmpty(errorMessage)) {
                                    resultCode = "error";
                                    resultMessage = errorMessage;
                                    break;
                                }
                            }
                        }
                    }
                }
            } else {
                resultMessage = "アトラクションセンターへのログイン中に不明な応答が返されました (" + response.StatusCode.ToString() + ")";
            }
            response.Close();
            // 認証結果分岐
            if(resultCode == "success") {
                // ログイン成功
                return;
            } else if(recursive) {
                // 再帰呼び出しを受けたうえで認証に失敗しているので、中止
                throw new AppException("認証に失敗しました(再帰認証)");
            } else if(resultCode == "idLock") {
                // IDロック用の認証コードを入力
                string authCode = null;
                if(IDLockAuth != null) {
                    IDLockAuthEventArgs e = new IDLockAuthEventArgs();
                    IDLockAuth(this, e);
                    authCode = e.AuthCode;
                }
                if(!string.IsNullOrEmpty(authCode)) {
                    // IDロック認証処理を行う
                    AuthIDLock(authCode);
                    // 再度ログイン処理を行う
                    Login(true);
                }
                // 再帰ログインに成功したら処理終了
                return;
            } else if(resultCode == "idLock2") {
                //throw new AppException("Not implemented!! / 電話認証によるIDロック機能への対応は完了していません");
                // 電話認証に必要なデータを取得し、画面に表示する
                IVRAuthRequestData authData = GetIVRAuthCode(resultMessage.Substring(resultMessage.IndexOf("?")));
                IVRAuthConfirmEventArgs confirm = new IVRAuthConfirmEventArgs();
                confirm.AuthCode = authData.AuthCode;
                IVRAuthConfirm(this, confirm);
                if(confirm.Cancel) {
                    // キャンセルボタンの処理
                    CancelIVRAuth(authData);
                    throw new AppException("認証がキャンセルされました");
                }
                // 電話認証を要求する
                authData = BeginIVRAuth(authData);
                // 電話認証の開始を通知する
                IVRAuthBeginCall(this, new EventArgs());
                IVRAuthStateData authState = null;
                int ivrRetry = 20; //適当な回数
                while(ivrRetry-- > 0) {
                    // 現在の状態を取得する
                    authState = GetIVRAuthState(authData);
                    if(authState.Finished) {
                        break;
                    }
                    // 初回以降はReferrerがいらないので、消しておく
                    if(authData.Referrer != null) {
                        authData.Referrer = null;
                    }
                    // 10秒間待機(METAタグ指定されているので)
                    System.Threading.Thread.Sleep(10000);
                }
                // 電話認証の終了を通知する
                IVRAuthEndCall(this, new EventArgs());
                if(ivrRetry < 0 || authState == null) {
                    throw new AppException("規定回数以内で電話認証処理が完了しませんでした");
                }
                if(!authState.Successful) {
                    throw new AppException("電話認証が正しく行えませんでした");
                }
                // 認証成功ならばIDロック用Cookieを取得する
                AccountData.IDLockCookie = GetIDLockCookie();
                if(AccountData.IDLockCookie == null) {
                    throw new AppException("IDロックCookieが取得できませんでした");
                }
                // 再度ログイン処理を行う
                Login(true);
                // 再帰ログインに成功したら処理終了
                return;
            } else if(resultCode == "imageauth") {
                // 画像認証が要求されたので、リダイレクト用パラメータからOTPを取得する
                string query = resultMessage.Substring(resultMessage.IndexOf("?"));
                NameValueCollection queries = HttpUtility.ParseQueryString(query);
                AuthImage(queries["otp"]);
                // 例外が発生しなかったならば、ログイン成功
                return;
            } else if(resultCode == "error") {
                // なにかのエラー
                throw new AppException(resultMessage);
            }
            // それ以外の応答は、よくわからない。
            throw new AppException(resultMessage);
        }

        /// <summary>
        /// IDロック認証処理
        /// </summary>
        /// <param name="authCode"></param>
        private void AuthIDLock(string authCode) {
            // とりあえず、ASP.NETのセッションキーとVIEWSTATEをもらってくる
            CurrentStateMessage = "IDロック認証ページからCookieを取得中";
            cookieContainer = new CookieContainer(); // Cookieを初期化
            HttpWebRequest request = CreateHttpWebRequest(URL_FRONT_SL_AUTH);
            HttpWebResponse response = (HttpWebResponse)request.GetResponse();
            Debug.WriteLine("AuthIDLock: response1=" + response.StatusCode);
            if(response.StatusCode != HttpStatusCode.OK) {
                response.Close();
                throw new AppException("IDロック認証ページでエラーが発生しました");
            }
            string viewStateString = null;
            using(StreamReader reader = new StreamReader(response.GetResponseStream(), Encoding.GetEncoding(response.CharacterSet))) {
                Regex regex = new Regex("^.*<input type=\"hidden\" name=\"__VIEWSTATE\" id=\"__VIEWSTATE\" value=\"(.+)\" />.*$");
                string line = null;
                Match match = null;
                while((line = reader.ReadLine()) != null) {
                    match = regex.Match(line);
                    if(match.Success) {
                        viewStateString = match.Groups[1].Value;
                        break;
                    }
                }
            }
            response.Close();
            if(viewStateString == null) {
                throw new AppException("IDロック認証ページからVIEWSTATEが取得できませんでした");
            }
            // IDロック認証の実行
            CurrentStateMessage = "IDロック認証中";
            request = CreateHttpWebRequest(URL_FRONT_SL_AUTH);
            request.Method = "POST";
            request.ContentType = "application/x-www-form-urlencoded";
            request.Referer = URL_FRONT_SL_AUTH;
            StringBuilder postData = new StringBuilder();
            postData.Append("__LASTFOCUS=");
            postData.Append("&__EVENTTARGET=");
            postData.Append("&__EVENTARGUMENT=");
            postData.Append("&__VIEWSTATE=");
            postData.Append(HttpUtility.UrlEncode(viewStateString));
            postData.Append("&ctl00%24ctl00%24MainContent%24TopContent%24attestKeyControl%24txtAttestKey=");
            postData.Append(HttpUtility.UrlEncode(authCode));
            postData.Append("&ctl00%24ctl00%24MainContent%24TopContent%24ibtNext.x=112");
            postData.Append("&ctl00%24ctl00%24MainContent%24TopContent%24ibtNext.y=12");
            byte[] postBytes = Encoding.ASCII.GetBytes(postData.ToString());
            request.ContentLength = postBytes.Length;
            using(Stream postStream = request.GetRequestStream()) {
                postStream.Write(postBytes, 0, postBytes.Length);
            }
            response = (HttpWebResponse)request.GetResponse();
            Debug.WriteLine("AuthIDLock: response2=" + response.StatusCode);
            if(response.StatusCode != HttpStatusCode.OK) {
                throw new AppException("IDロック認証中に不明な応答が返されました (" + response.StatusCode.ToString() + ")");
            }
            response.Close();
            // Cookieを取得する
            AccountData.IDLockCookie = GetIDLockCookie();
            if(AccountData.IDLockCookie == null) {
                throw new AppException("IDロックCookieが取得できませんでした");
            }
        }

        /// <summary>
        /// 画像認証
        /// </summary>
        /// <param name="otp"></param>
        private void AuthImage(string otp) {
            // 画像認証用のデータを取得
            StringBuilder authUrl = new StringBuilder();
            authUrl.Append(URL_FRONT_IA_AUTH);
            authUrl.Append("?otp=");
            authUrl.Append(otp);
            authUrl.Append("&.goeReturnUrl=/front/member/center.aspx");
            authUrl.Append("&.goeBackUrl=/front/member/center.aspx");
            CurrentStateMessage = "画像認証ページからデータを取得中";
            HttpWebRequest request = CreateHttpWebRequest(authUrl.ToString());
            HttpWebResponse response = (HttpWebResponse)request.GetResponse();
            Debug.WriteLine("AuthImage: response1=" + response.StatusCode);
            if(response.StatusCode != HttpStatusCode.OK) {
                response.Close();
                throw new AppException("画像認証ページの取得中にエラーが発生しました");
            }
            string viewStateString = null;
            Encoding enc = Encoding.GetEncoding(response.CharacterSet);
            using(StreamReader reader = new StreamReader(response.GetResponseStream(), enc)) {
                Regex regex = new Regex("^.*<input type=\"hidden\" name=\"__VIEWSTATE\" id=\"__VIEWSTATE\" value=\"(.+)\" />.*$");
                string line = null;
                Match match = null;
                while((line = reader.ReadLine()) != null) {
                    match = regex.Match(line);
                    if(match.Success) {
                        viewStateString = match.Groups[1].Value;
                        break;
                    }
                }
            }
            response.Close();
            if(viewStateString == null) {
                throw new AppException("画像認証ページからVIEWSTATEが取得できませんでした");
            }
            // 画像の取得
            request = CreateHttpWebRequest(URL_FRONT_IA_IMAGE);
            response = (HttpWebResponse)request.GetResponse();
            Debug.WriteLine("AuthImage: response2=" + response.StatusCode);
            Image captchaImage = Image.FromStream(response.GetResponseStream());
            // 画像を表示し、入力を促す
            string captchaText = null;
            if(ImageAuth != null) {
                ImageAuthEventArgs e = new ImageAuthEventArgs();
                e.CaptchaImage = captchaImage;
                e.SaveAuthState = AccountData.SaveImageAuthCookie;
                ImageAuth(this, e);
                captchaText = e.CaptchaText;
                AccountData.SaveImageAuthCookie = e.SaveAuthState;
            }
            //// 画像認証の実行
            CurrentStateMessage = "画像認証中";
            request = CreateHttpWebRequest(authUrl.ToString());
            request.Method = "POST";
            request.ContentType = "application/x-www-form-urlencoded";
            request.Referer = authUrl.ToString();
            StringBuilder postData = new StringBuilder();
            postData.Append("__LASTFOCUS=");
            postData.Append("&__EVENTTARGET=");
            postData.Append("&__EVENTARGUMENT=");
            postData.Append("&__VIEWSTATE=");
            postData.Append(HttpUtility.UrlEncode(viewStateString));
            postData.Append("&ctl00%24ctl00%24MainContent%24TopContent%24captchaControlAjax%24txtCaptcha=");
            postData.Append(HttpUtility.UrlEncode(captchaText, enc));
            if(AccountData.SaveImageAuthCookie) {
                postData.Append("&ctl00%24ctl00%24MainContent%24TopContent%24chbSave=on");
            }
            postData.Append("&ctl00%24ctl00%24MainContent%24TopContent%24txt=");
            postData.Append("&ctl00%24ctl00%24MainContent%24TopContent%24btnNext=");
            byte[] postBytes = Encoding.ASCII.GetBytes(postData.ToString());
            request.ContentLength = postBytes.Length;
            using(Stream postStream = request.GetRequestStream()) {
                postStream.Write(postBytes, 0, postBytes.Length);
            }
            response = (HttpWebResponse)request.GetResponse();
            response.Close();
            Debug.WriteLine("AuthImage: response3=" + response.StatusCode);
            if(response.StatusCode == HttpStatusCode.Found) {
                if(AccountData.SaveImageAuthCookie) {
                    // Cookieを取得する
                    AccountData.ImageAuthCookie = GetImageAuthCookie();
                    if(AccountData.ImageAuthCookie == null) {
                        throw new AppException("画像認証結果のCookieが取得できませんでした");
                    }
                }
            } else if(response.StatusCode == HttpStatusCode.OK) {
                throw new AppException("画像認証に失敗しました (ID/パスワード/画像認証の何れかが間違っています)");
            } else {
                throw new AppException("画像認証中に不明な応答が返されました (" + response.StatusCode.ToString() + ")");
            }
        }

        /// <summary>
        /// IDロック機能用のCookieデータを取得する
        /// </summary>
        /// <returns></returns>
        private Cookie GetIDLockCookie() {
            if(cookieContainer == null) {
                return null;
            }
            CookieCollection cookies = cookieContainer.GetCookies(new Uri(URL_FRONT_LOGIN));
            Cookie cookie = cookies["GSL"];
            if(cookie == null) {
                return null;
            }
            if(cookie.Expired) {
                return null;
            }
            return cookie;
        }

        /// <summary>
        /// 画像認証状態保持用のCookieデータを取得する
        /// </summary>
        /// <returns></returns>
        private Cookie GetImageAuthCookie() {
            if(cookieContainer == null) {
                Debug.WriteLine("GetImageAuthCookie: no container");
                return null;
            }
            CookieCollection cookies = cookieContainer.GetCookies(new Uri(URL_FRONT_LOGIN));
            Cookie cookie = null;
            foreach(Cookie ck in cookies) {
#if DEBUG
                Debug.WriteLine("GetImageAuthCookie: search: [" + ck.Name + "]");
#endif
                if(ck.Name.StartsWith("GHLI") && ck.Name.Length == 36 && ck.Expired == false) {
                    // ドメイン部を書き換えておく
                    // (なんでサブドメインに対して送信されない？CookieContainerの仕様かな。。)
                    ck.Domain = (new Uri(URL_FRONT_LOGIN)).Host;
                    cookie = ck;
                    break;
                }
            }
#if DEBUG
            if(cookie == null) {
                Debug.WriteLine("GetImageAuthCookie: cookie not found");
            } else {
                StringBuilder sb = new StringBuilder();
                sb.Append("GetImageAuthCookie: cookie found:\r\n");
                sb.Append(" Name=" + cookie.Name + "\r\n");
                sb.Append(" Value=" + cookie.Value + "\r\n");
                sb.Append(" Domain=" + cookie.Domain + "\r\n");
                sb.Append(" Port=" + cookie.Port + "\r\n");
                sb.Append(" Path=" + cookie.Path + "\r\n");
                sb.Append(" Secure=" + cookie.Secure + "\r\n");
                sb.Append(" Expires=" + cookie.Expires + "\r\n");
                sb.Append(" Expired=" + cookie.Expired + "\r\n");
                Debug.WriteLine(sb.ToString());
            }
#endif
            return cookie;
        }

        /// <summary>
        /// 電話認証で入力するべきコード(数列)を取得する
        /// </summary>
        /// <param name="query"></param>
        /// <returns></returns>
        private IVRAuthRequestData GetIVRAuthCode(string query) {
            // 全画面リダイレクトページを取得 (たぶん不要)
            StringBuilder redirUrl = new StringBuilder();
            redirUrl.Append(URL_FRONT_IL_REDIRECT);
            redirUrl.Append(query);
            HttpWebRequest request = CreateHttpWebRequest(redirUrl.ToString());
            HttpWebResponse response = (HttpWebResponse)request.GetResponse();
            Debug.WriteLine("GetIVRAuthCode: response1=" + response.StatusCode);
            response.Close();
            if(response.StatusCode != HttpStatusCode.OK) {
                throw new AppException("全画面リダイレクトページのGETで不明な応答が返されました (" + response.StatusCode.ToString() + ")");
            }
            // 本人確認実行確認ページを取得 (たぶん不要)
            StringBuilder verifyUrl = new StringBuilder();
            verifyUrl.Append(URL_FRONT_SL_VERIFY);
            verifyUrl.Append(query);
            request = CreateHttpWebRequest(verifyUrl.ToString());
            request.Referer = redirUrl.ToString();
            response = (HttpWebResponse)request.GetResponse();
            Debug.WriteLine("GetIVRAuthCode: response2=" + response.StatusCode);
            if(response.StatusCode != HttpStatusCode.OK) {
                response.Close();
                throw new AppException("本人確認実行確認ページのGETで不明な応答が返されました (" + response.StatusCode.ToString() + ")");
            }
            string viewStateString = null;
            using(StreamReader reader = new StreamReader(response.GetResponseStream(), Encoding.GetEncoding(response.CharacterSet))) {
                Regex regex = new Regex("^.*<input type=\"hidden\" name=\"__VIEWSTATE\" id=\"__VIEWSTATE\" value=\"(.+)\" />.*$");
                string line = null;
                Match match = null;
                while((line = reader.ReadLine()) != null) {
                    match = regex.Match(line);
                    if(match.Success) {
                        viewStateString = match.Groups[1].Value;
                        break;
                    }
                }
            }
            response.Close();
            if(viewStateString == null) {
                throw new AppException("本人確認実行確認ページからVIEWSTATEが取得できませんでした");
            }
            // 本人確認実行確認ページで次画面遷移 (たぶん不要)
            request = CreateHttpWebRequest(verifyUrl.ToString());
            request.Method = "POST";
            request.ContentType = "application/x-www-form-urlencoded";
            request.Referer = verifyUrl.ToString();
            StringBuilder postData = new StringBuilder();
            postData.Append("__VIEWSTATE=");
            postData.Append(HttpUtility.UrlEncode(viewStateString));
            postData.Append("&ctl00$ctl00$MainContent$TopContent$ibtNext.x=122");
            postData.Append("&ctl00$ctl00$MainContent$TopContent$ibtNext.y=15");
            byte[] postBytes = Encoding.ASCII.GetBytes(postData.ToString());
            request.ContentLength = postBytes.Length;
            using(Stream postStream = request.GetRequestStream()) {
                postStream.Write(postBytes, 0, postBytes.Length);
            }
            response = (HttpWebResponse)request.GetResponse();
            Debug.WriteLine("GetIVRAuthCode: response3=" + response.StatusCode);
            if(response.StatusCode != HttpStatusCode.Found) {
                throw new AppException("本人確認実行確認ページへのPOSTで不明な応答が返されました (" + response.StatusCode.ToString() + ")");
            }
            string redir = response.Headers["Location"];
            if(!redir.StartsWith(URL_FRONT_SL_PRENOTEIVR_PATH)) {
                throw new AppException("本人確認実行確認ページへのPOSTで不明なリダイレクトが返されました (" + redir + ")");
            }
            response.Close();
            // 注意事項ページデータの取得
            StringBuilder prenoteUrl = new StringBuilder();
            prenoteUrl.Append(URL_FRONT_SL_PRENOTEIVR);
            prenoteUrl.Append(redir.Substring(redir.IndexOf("?")));
            request = CreateHttpWebRequest(prenoteUrl.ToString());
            request.Referer = verifyUrl.ToString();
            response = (HttpWebResponse)request.GetResponse();
            Debug.WriteLine("GetIVRAuthCode: response4=" + response.StatusCode);
            if(response.StatusCode != HttpStatusCode.OK) {
                response.Close();
                throw new AppException("注意事項ページのGETで不明な応答が返されました (" + response.StatusCode.ToString() + ")");
            }
            viewStateString = null;
            string authCode = null;
            using(StreamReader reader = new StreamReader(response.GetResponseStream(), Encoding.GetEncoding(response.CharacterSet))) {
                Regex regex1 = new Regex("^.*<input type=\"hidden\" name=\"__VIEWSTATE\" id=\"__VIEWSTATE\" value=\"(.+)\" />.*$");
                Regex regex2 = new Regex("^.+<span id=\"MainContent_TopContent_labAuthCode\">(.+)</span>.*$");
                string line = null;
                Match match = null;
                while((line = reader.ReadLine()) != null) {
                    if(viewStateString == null) {
                        match = regex1.Match(line);
                        if(match.Success) {
                            viewStateString = match.Groups[1].Value;
                        }
                    }
                    if(authCode == null) {
                        match = regex2.Match(line);
                        if(match.Success) {
                            authCode = match.Groups[1].Value;
                        }
                    }
                }
            }
            response.Close();
            if(viewStateString == null) {
                throw new AppException("注意事項ページからVIEWSTATEが取得できませんでした");
            }
            if(authCode == null) {
                throw new AppException("注意事項ページから認証コードが取得できませんでした");
            }
            // 集めたデータを返す
            IVRAuthRequestData authData = new IVRAuthRequestData();
            authData.Url = prenoteUrl.ToString();
            authData.Referrer = verifyUrl.ToString();
            authData.ViewState = viewStateString;
            authData.AuthCode = authCode;
            return authData;
        }

        /// <summary>
        /// 電話認証開始要求
        /// </summary>
        /// <param name="authData"></param>
        private IVRAuthRequestData BeginIVRAuth(IVRAuthRequestData authData) {
            // 注意事項ページで次画面遷移
            HttpWebRequest request = CreateHttpWebRequest(authData.Url);
            request.Method = "POST";
            request.ContentType = "application/x-www-form-urlencoded";
            request.Referer = authData.Referrer;
            StringBuilder postData = new StringBuilder();
            postData.Append("__VIEWSTATE=");
            postData.Append(HttpUtility.UrlEncode(authData.ViewState));
            postData.Append("&ctl00$ctl00$MainContent$TopContent$ibtNext.x=89");
            postData.Append("&ctl00$ctl00$MainContent$TopContent$ibtNext.y=15");
            byte[] postBytes = Encoding.ASCII.GetBytes(postData.ToString());
            request.ContentLength = postBytes.Length;
            using(Stream postStream = request.GetRequestStream()) {
                postStream.Write(postBytes, 0, postBytes.Length);
            }
            HttpWebResponse response = (HttpWebResponse)request.GetResponse();
            Debug.WriteLine("BeginIVRAuth: response1=" + response.StatusCode);
            if(response.StatusCode != HttpStatusCode.Found) {
                throw new AppException("注意事項ページへのPOST(電話認証)で不明な応答が返されました (" + response.StatusCode.ToString() + ")");
            }
            string redir = response.Headers["Location"];
            if(!redir.StartsWith(URL_FRONT_SL_IVRAUTH_PATH)) {
                throw new AppException("注意事項ページへのPOST(電話認証)で不明なリダイレクトが返されました (" + redir + ")");
            }
            response.Close();
            // 電話認証ページのURLを組み立て
            StringBuilder authUrl = new StringBuilder();
            authUrl.Append(URL_FRONT_SL_IVRAUTH);
            authUrl.Append(redir.Substring(redir.IndexOf("?")));
            // 新しい認証データを生成して返す
            IVRAuthRequestData nextReq = new IVRAuthRequestData();
            nextReq.AuthCode = authData.AuthCode;
            nextReq.Url = authUrl.ToString();
            nextReq.Referrer = null;
            return nextReq;
        }

        /// <summary>
        /// 電話認証状態取得
        /// </summary>
        /// <param name="authData"></param>
        /// <returns></returns>
        private IVRAuthStateData GetIVRAuthState(IVRAuthRequestData authData) {
            IVRAuthStateData state = new IVRAuthStateData();
            state.Finished = false;
            state.Successful = false;
            // 認証状況ページを取得
            HttpWebRequest request = CreateHttpWebRequest(authData.Url);
            if(!string.IsNullOrEmpty(authData.Referrer)) {
                request.Referer = authData.Referrer;
            }
            HttpWebResponse response = (HttpWebResponse)request.GetResponse();
            Debug.WriteLine("GetIVRAuthState: response1=" + response.StatusCode);
            if(response.StatusCode == HttpStatusCode.Found) {
                // 302が返ってきた場合は認証エラー
                response.Close();
                if(response.Headers["Location"].StartsWith(URL_FRONT_INFO_PATH)) {
                    throw new AppException("電話認証に失敗しました");
                }
                throw new AppException("電話認証で予期せぬリダイレクトを受け取りました (" + response.Headers["Location"] + ")");
            } else if(response.StatusCode != HttpStatusCode.OK) {
                // 302/200以外の応答は異常
                response.Close();
                throw new AppException("電話認証状態ページのGETで不明な応答が返されました (" + response.StatusCode.ToString() + ")");
            }
            string viewStateString = null;
            using(StreamReader reader = new StreamReader(response.GetResponseStream(), Encoding.GetEncoding(response.CharacterSet))) {
                Regex regex = new Regex("^.*<input type=\"hidden\" name=\"__VIEWSTATE\" id=\"__VIEWSTATE\" value=\"(.+)\" />.*$");
                Match match = null;
                string checkText1 = "■電話認証中です■<br />";
                string checkText2 = "IDロックCookie登録完了しました。<br />";
                string line = null;
                while((line = reader.ReadLine()) != null) {
                    if(line == checkText1) {
                        // まだ認証を実行している最中である
                        break;
                    }
                    if(line == checkText2) {
                        // 正常終了
                        state.Finished = true;
                        state.Successful = true;
                        break;
                    }
                    if(viewStateString == null) {
                        match = regex.Match(line);
                        if(match.Success) {
                            viewStateString = match.Groups[1].Value;
                        }
                    }
                }
            }
            response.Close();
            if(state.Finished && state.Successful) {
                // 完了ページを読み込む
                request = CreateHttpWebRequest(URL_FRONT_SL_FINISH + authData.Url.Substring(authData.Url.IndexOf("?")));
                request.Method = "POST";
                request.ContentType = "application/x-www-form-urlencoded";
                request.Referer = authData.Url;
                StringBuilder postData = new StringBuilder();
                postData.Append("__VIEWSTATE=");
                postData.Append(HttpUtility.UrlEncode(viewStateString));
                postData.Append("&ctl00$ctl00$MainContent$TopContent$ibtNext.x=135");
                postData.Append("&ctl00$ctl00$MainContent$TopContent$ibtNext.y=9");
                byte[] postBytes = Encoding.ASCII.GetBytes(postData.ToString());
                request.ContentLength = postBytes.Length;
                using(Stream postStream = request.GetRequestStream()) {
                    postStream.Write(postBytes, 0, postBytes.Length);
                }
                response = (HttpWebResponse)request.GetResponse();
                Debug.WriteLine("GetIVRAuthState: response2=" + response.StatusCode);
                response.Close();
                if(response.StatusCode != HttpStatusCode.Found) {
                    throw new AppException("電話認証完了ページへのPOSTで不明な応答が返されました (" + response.StatusCode.ToString() + ")");
                }
            }
            return state;
        }

        /// <summary>
        /// 電話認証処理キャンセル
        /// </summary>
        /// <param name="authData"></param>
        private void CancelIVRAuth(IVRAuthRequestData authData) {
            // 注意事項ページで次画面遷移
            HttpWebRequest request = CreateHttpWebRequest(authData.Url);
            request.Method = "POST";
            request.ContentType = "application/x-www-form-urlencoded";
            request.Referer = authData.Referrer;
            StringBuilder postData = new StringBuilder();
            postData.Append("__VIEWSTATE=");
            postData.Append(HttpUtility.UrlEncode(authData.ViewState));
            postData.Append("&ctl00$ctl00$MainContent$TopContent$ibtBack.x=59");
            postData.Append("&ctl00$ctl00$MainContent$TopContent$ibtBack.y=11");
            byte[] postBytes = Encoding.ASCII.GetBytes(postData.ToString());
            request.ContentLength = postBytes.Length;
            using(Stream postStream = request.GetRequestStream()) {
                postStream.Write(postBytes, 0, postBytes.Length);
            }
            HttpWebResponse response = (HttpWebResponse)request.GetResponse();
            Debug.WriteLine("CancelIVRAuth: response=" + response.StatusCode);
            response.Close();
            if(response.StatusCode != HttpStatusCode.Found) {
                throw new AppException("注意事項ページへのPOST(キャンセル)で不明な応答が返されました (" + response.StatusCode.ToString() + ")");
            }
        }

        /// <summary>
        /// ハンゲームへのログインが必要か判定
        /// </summary>
        /// <returns></returns>
        private bool IsHanGameLoginRequired() {
            if(cookieContainer == null) {
                return true;
            }
            CookieCollection cookies = cookieContainer.GetCookies(new Uri(URL_HG_RAGNAROK));
            Cookie loginCookie = cookies["login"];
            if(loginCookie == null) {
                return true;
            }
            if(loginCookie.Value == null) {
                return true;
            }
            if(loginCookie.Value == "null") {
                return true;
            }
            return false;
        }

        /// <summary>
        /// ハンゲーム用ログイン処理
        /// </summary>
        private void LoginHanGame() {
            // Cookieを取得
            CurrentStateMessage = "ハンゲームよりCookieを取得中";
            cookieContainer = new CookieContainer(); // Cookieを初期化
            HttpWebRequest request = CreateHttpWebRequest(URL_HG_RAGNAROK);
            HttpWebResponse response = (HttpWebResponse)request.GetResponse();
            Debug.WriteLine("LoginHanGame: response1=" + response.StatusCode);
            if(response.StatusCode != HttpStatusCode.OK) {
                response.Close();
                throw new AppException("ハンゲームへ接続中にエラーが発生しました");
            }
            response.Close();
            // ログインに必要な情報を投げつける
            string userId = AccountData.Id.Substring(3);
            if(userId.Length > 20) {
                userId = userId.Substring(0, 20);
            }
            byte[] userIdBytes = Encoding.Unicode.GetBytes(userId);
            Encoding sjisEncode = Encoding.GetEncoding("Shift_JIS");
            userIdBytes = Encoding.Convert(Encoding.Unicode, sjisEncode, userIdBytes);
            CurrentStateMessage = "ハンゲームへ " + userId + " でログイン中";
            request = CreateHttpWebRequest(URL_HG_LOGIN);
            request.Method = "POST";
            request.ContentType = "application/x-www-form-urlencoded";
            request.Referer = URL_HG_RAGNAROK;
            StringBuilder postData = new StringBuilder();
            postData.Append("nxtUrl=http%3A%2F%2Fragnarokonline.hangame.co.jp%2Findex.nhn");
            postData.Append("&strmemberid=");
            postData.Append(HttpUtility.UrlEncode(userIdBytes));
            postData.Append("&strpassword=");
            postData.Append(HttpUtility.UrlEncode(AccountData.Password.Substring(0, (AccountData.Password.Length < 16) ? AccountData.Password.Length : 16)));
            postData.Append("&chkSSLLogin=on");
            byte[] postBytes = Encoding.ASCII.GetBytes(postData.ToString());
            request.ContentLength = postBytes.Length;
            using(Stream postStream = request.GetRequestStream()) {
                postStream.Write(postBytes, 0, postBytes.Length);
            }
            response = (HttpWebResponse)request.GetResponse();
            response.Close();
            Debug.WriteLine("LoginHanGame: response2=" + response.StatusCode);
            if(response.StatusCode != HttpStatusCode.OK) {
                throw new AppException("ハンゲームへのログイン処理でエラーが発生しました");
            }
            // ログインできたか確認
            if(IsHanGameLoginRequired()) {
                throw new AppException("ハンゲームへログインできませんでした。ID/パスワードを確認してください。");
            }
            CurrentStateMessage = "ハンゲームへのログインが完了";
            Debug.WriteLine("LoginHanGame: login successful");
            return;
        }


        /// <summary>
        /// アトラクションセンターのログアウト処理を行う
        /// </summary>
        public void Logout() {
            if(!this.IsLoggedIn) {
                return;
            }
            CurrentStateMessage = "アトラクションセンターからログアウト中";
            // ログアウトボタンを押したのと同じ状況を作る
            HttpWebRequest request = CreateHttpWebRequest(URL_FRONT_WEBGS);
            request.Method = "POST";
            request.ContentType = "application/x-www-form-urlencoded";
            request.Referer = URL_FRONT_WEBGS;
            StringBuilder postData = new StringBuilder();
            postData.Append("__VIEWSTATE=");
            postData.Append(HttpUtility.UrlEncode(webGSViewState));
            postData.Append("&ddlCPID=select");
            postData.Append("&logout=");
            byte[] postBytes = Encoding.ASCII.GetBytes(postData.ToString());
            request.ContentLength = postBytes.Length;
            using(Stream postStream = request.GetRequestStream()) {
                postStream.Write(postBytes, 0, postBytes.Length);
            }
            request.AllowAutoRedirect = true;
            HttpWebResponse response = (HttpWebResponse)request.GetResponse();
            response.Close();
            Debug.WriteLine("Logout: response=" + response.StatusCode);
            webGSViewState = null;
            CurrentStateMessage = "アトラクションセンターからのログアウトが完了";
            Debug.WriteLine("Logout: successful");
            return;
        }


        /// <summary>
        /// アトラクションIDの一覧を取得する
        /// </summary>
        public void GetAttractionAccountData() {
            if(AccountData == null) {
                throw new AppException("ログイン用のデータが設定されていません");
            }
            if(string.IsNullOrEmpty(AccountData.Id)) {
                throw new AppException("ログイン用のIDが設定されていません");
            }
            // ハンゲーム用IDだったら、そちらへ処理をまわす
            if(AccountData.Id.StartsWith("HG:")) {
                GetHanGameAttractionAccountData();
                return;
            }
            bool retry = false;
            do {
                CurrentStateMessage = "アトラクションID情報の取得を開始";
                HttpWebRequest request = CreateHttpWebRequest(URL_FRONT_WEBGS);
                // request.Referer = URL_FRONT_LOGIN + "?ReturnUrl=" + HttpUtility.UrlEncode(URL_FRONT_WEBGS_FILE);
                HttpWebResponse response = (HttpWebResponse)request.GetResponse();
                if(response.StatusCode == HttpStatusCode.OK) {
                    // Web起動用画面が普通に取得できた場合
                    retry = false;
                    Debug.WriteLine("GetAttractionAccountData: response=OK");
                    // TODO: 毎回リストを作りなおす必要があるか？ 既存のリストとの照らし合わせで追加削除するほうが。
                    // 今のところ考え直すのが面倒だからそのままで...
                    CurrentStateMessage = "アトラクションID情報の解析中";
                    List<AttractionAccountData> attAccountList = new List<AttractionAccountData>();
                    using(StreamReader reader = new StreamReader(response.GetResponseStream(), Encoding.GetEncoding(response.CharacterSet))) {
                        string line = null;
                        Regex vsRegex = new Regex("^.*<input type=\"hidden\" name=\"__VIEWSTATE\" id=\"__VIEWSTATE\" value=\"(.+)\" />.*$");
                        Match vsMatch = null;
                        webGSViewState = null;
                        string nextIdDiv = null;
                        int nextIdIndex = 0;
                        while((line = reader.ReadLine()) != null) {
                            if(string.IsNullOrEmpty(line)) {
                                continue;
                            }
                            line = line.Trim();
                            // ViewStateが取得できるか試す
                            if(webGSViewState == null) {
                                vsMatch = vsRegex.Match(line);
                                if(vsMatch.Success) {
                                    webGSViewState = vsMatch.Groups[1].Value;
                                }
                            }
                            // まず、SIIDを取得できる行を探し出す
                            if(line == "<option value=\"select\">ゲームアカウントを選択してください</option>") {
                                // 次の行から</select>までがIDリスト
                                Regex r = new Regex("^<option value=\"(.+)\">(.+)\\((.+)\\)</option>$");
                                Match m = null;
                                while((line = reader.ReadLine()) != null) {
                                    if(string.IsNullOrEmpty(line)) {
                                        continue;
                                    }
                                    line = line.Trim();
                                    if(line == "</select>") {
                                        // IDリスト終了
                                        break;
                                    }
                                    m = r.Match(line);
                                    if(m.Success) {
                                        AttractionAccountData attAccount = new AttractionAccountData();
                                        attAccount.Id = m.Groups[3].Value;
                                        attAccount.Password = "";
                                        attAccount.InternalId = m.Groups[1].Value;
                                        attAccount.Name = m.Groups[2].Value;
                                        attAccount.ExpiryDateString = "";
                                        attAccountList.Add(attAccount);
                                        if(nextIdDiv == null) {
                                            nextIdDiv = "<div id=\"" + attAccount.InternalId + "\">";
                                            nextIdIndex = 0;
                                        }
                                    }
                                }
                            }
                            // IDからその他の情報を取得する
                            if(line == nextIdDiv) {
                                AttractionAccountData attAccount = attAccountList[nextIdIndex];
                                // 次の行は有効期限
                                line = reader.ReadLine();
                                Regex r = new Regex("^.*<p>([^<]+)<br />.*$");
                                Match m = r.Match(line);
                                if(m.Success) {
                                    attAccount.ExpiryDateString = m.Groups[1].Value;
                                }
                                // 次のID情報取得準備
                                nextIdIndex++;
                                if(nextIdIndex < attAccountList.Count) {
                                    nextIdDiv = "<div id=\"" + attAccountList[nextIdIndex].InternalId + "\">";
                                } else {
                                    nextIdDiv = null;
                                }
                            }

                        }
                    }
                    response.Close();
                    AccountData.AttractionAccounts = attAccountList;
                } else {
                    // Web起動用画面が取得できなかった場合
                    CurrentStateMessage = "アトラクションID情報の取得に失敗";
                    string location = response.Headers["Location"];
                    response.Close();
                    if(retry) {
                        // 既に一度リトライされているので、ログインしてもダメであったということ？
                        StringBuilder msg = new StringBuilder();
                        msg.Append("アトラクションID情報が取得できませんでした(リトライエラー)\r\n");
                        msg.Append("ステータス=");
                        msg.Append(response.StatusCode.ToString());
                        msg.Append("\r\n");
                        if(!string.IsNullOrEmpty(location)) {
                            msg.Append("リダイレクト=");
                            msg.Append(location);
                            msg.Append("\r\n");
                        }
                        throw new AppException(msg.ToString());
                    }
                    if(response.StatusCode == HttpStatusCode.Found) {
                        if(string.IsNullOrEmpty(location)) {
                            throw new AppException("アトラクションID情報の取得中に、不明なリダイレクトが発生しました (URL=null)");
                        }
                        if (!location.StartsWith(URL_FRONT_LOGIN_PATH, StringComparison.OrdinalIgnoreCase)
                            && !location.StartsWith(URL_FRONT_LOGIN, StringComparison.OrdinalIgnoreCase)) {
                            throw new AppException("アトラクションID情報の取得中に、不明なリダイレクトが発生しました (URL=" + location + ")");
                        }
                        // ログインが必要なご様子。処理を行ってリトライ？
                        CurrentStateMessage = "アトラクションセンターへのログインを試行";
                        Debug.WriteLine("GetAttractionAccountData: require Login()");
                        Login();
                        retry = true;
                    } else {
                        // 想定していないコードが返ってきた
                        StringBuilder msg = new StringBuilder();
                        msg.Append("アトラクションID情報が取得できませんでした(想定外ステータス)\r\n");
                        msg.Append("ステータス=");
                        msg.Append(response.StatusCode.ToString());
                        msg.Append("\r\n");
                        if(!string.IsNullOrEmpty(location)) {
                            msg.Append("リダイレクト=");
                            msg.Append(location);
                            msg.Append("\r\n");
                        }
                        throw new AppException(msg.ToString());
                    }
                }
            } while(retry);
            CurrentStateMessage = "アトラクションID情報の取得が完了";
            return;
        }


        /// <summary>
        /// ハンゲーム用アトラクション情報取得
        /// </summary>
        private void GetHanGameAttractionAccountData() {
            // ログインを試みる
            if(IsHanGameLoginRequired()) {
                LoginHanGame();
            }
            // ハンゲームの場合、アトラクションIDは1つしか存在しないため、固定でデータを作成する
            AttractionAccountData attAccount = new AttractionAccountData();
            attAccount.Id = "RagnarokOnline";
            attAccount.InternalId = "HG:RagnarokOnline";
            attAccount.Password = "";
            attAccount.Name = "ハンゲーム RagnarokOnline";
            if(AccountData.AttractionAccounts == null) {
                AccountData.AttractionAccounts = new List<AttractionAccountData>();
            } else {
                AccountData.AttractionAccounts.Clear();
            }
            AccountData.AttractionAccounts.Add(attAccount);
            return;
        }


        /// <summary>
        /// パスワードを取得する
        /// </summary>
        /// <param name="attractionInternalId">アトラクション内部ID(アトラクションセンターでSIIDパラメータとして扱われているもの)</param>
        /// <param name="isUrdrWorld">Urdrクライアント用のリクエストの場合、true</param>
        /// <remarks>
        /// 既にアトラクションセンターへログインしていることが前提となる。
        /// ログインしていない場合はリダイレクトを受けて例外が発生すると思われる。
        /// 取得したパスワードは、処理対象アカウントデータの中へ格納される。
        /// </remarks>
        public void GetOneTimePassword(string attractionInternalId, bool isUrdrWorld) {
            if(string.IsNullOrEmpty(attractionInternalId)) {
                throw new AppException("パスワード取得対象アカウントが指定されていません");
            }
            Debug.WriteLine("attractionInternalId=" + attractionInternalId);
            // TODO: パスワードをアカウントデータの中へ格納し直す必要があるか？ 考え直し
            // 対象のアカウントが存在しているか
            AttractionAccountData targetData = null;
            foreach(AttractionAccountData data in AccountData.AttractionAccounts) {
                if(data.InternalId == attractionInternalId) {
                    targetData = data;
                    break;
                }
            }
            if(targetData == null) {
                throw new AppException("指定されたアカウントの情報が処理対象リストに含まれていません");
            }
            // ハンゲームアカウントの場合は、別処理
            if(attractionInternalId.StartsWith("HG:")) {
                GetHanGameOneTimePassword(targetData);
                return;
            }
            // Webページからデータを奪ってくる
            CurrentStateMessage = "アトラクションパスワードの取得を開始";
            StringBuilder url = new StringBuilder();
            url.Append(URL_FRONT_WEBGS);
            url.Append(@"?SIID=");
            url.Append(attractionInternalId);
            if(isUrdrWorld) {
                url.Append("&world=urdr");
            }
            HttpWebRequest request = CreateHttpWebRequest(url.ToString());
            request.Referer = URL_FRONT_WEBGS;
            HttpWebResponse response = (HttpWebResponse)request.GetResponse();
            if(response.StatusCode != HttpStatusCode.OK) {
                response.Close();
                if(response.StatusCode != HttpStatusCode.Found) {
                    // リダイレクト応答以外は何かの異常と思われるので処理を中断する
                    throw new AppException("アトラクションパスワードの取得中に予期せぬ応答が返されました(1) [" + response.StatusCode.ToString() + "]");
                }
                string location = response.Headers["Location"];
                if(string.IsNullOrEmpty(location)) {
                    throw new AppException("アトラクションパスワードの取得中に不明なリダイレクトが発生しました (URL=null)");
                }
                if(!location.StartsWith(URL_FRONT_LAUNCH_REDIRECT_PATH, StringComparison.OrdinalIgnoreCase)) {
                    throw new AppException("アトラクションパスワードの取得中に不明なリダイレクトが発生しました (URL=" + location + ")");
                }
                // ゲーム起動案内用の画面へのリダイレクトなので、無視してActiveX呼び出し用の画面へ遷移する
                request = CreateHttpWebRequest(string.Format("{0}?SIID={1}&world={2}", URL_FRONT_LAUNCH, attractionInternalId, isUrdrWorld ? "urdr" : ""));
                request.Referer = string.Format("{0}?SIID={1}&world={2}", URL_FRONT_TERM, attractionInternalId, isUrdrWorld ? "urdr" : "");
                response = (HttpWebResponse)request.GetResponse();
                if(response.StatusCode != HttpStatusCode.OK) {
                    response.Close();
                    throw new AppException("アトラクションパスワードの取得中に予期せぬ応答が返されました(2) [" + response.StatusCode.ToString() + "]");
                }
            }
            CurrentStateMessage = "アトラクションパスワード情報の解析中";
            Dictionary<string, string> paramTable = new Dictionary<string, string>();
            using(StreamReader reader = new StreamReader(response.GetResponseStream(), Encoding.GetEncoding(response.CharacterSet))) {
                string line = null;
                Regex paramRegex = new Regex("^.*GameStartAsync\\('([^']+)'");
                Match match = null;
                while((line = reader.ReadLine()) != null) {
                    match = paramRegex.Match(line);
                    if(match.Success) {
                        paramTable.Add("gameStartParam", match.Groups[1].Value);
                    }
                }
            }
            response.Close();
            if(!paramTable.ContainsKey("gameStartParam")) {
                throw new AppException("アトラクションパスワードの取得に失敗しました");
            }
            targetData.Password = paramTable["gameStartParam"];
            CurrentStateMessage = "アトラクションパスワードの取得が完了";
            return;
        }



        /// <summary>
        /// ハンゲーム用パスワード取得処理
        /// </summary>
        /// <param name="attAccount"></param>
        public void GetHanGameOneTimePassword(AttractionAccountData targetData) {
            // ハンゲーム側からガンホーの起動画面まで、リダイレクトしてもらう
            CurrentStateMessage = "アトラクションパスワード(HG)の取得を開始";
            HttpWebRequest request = CreateHttpWebRequest(URL_HG_WEBGS);
            request.AllowAutoRedirect = true;
            request.Referer = URL_HG_RAGNAROK;
            HttpWebResponse response = (HttpWebResponse)request.GetResponse();
            if(response.StatusCode != HttpStatusCode.OK) {
                response.Close();
                throw new AppException("アトラクションパスワード(HG)の取得中に予期せぬ応答が返されました(1) [" + response.StatusCode.ToString() + "]");
            }
            Uri checkUri = new Uri(URL_FRONT_WEBGS_HG);
            if(response.ResponseUri.Host != checkUri.Host) {
                response.Close();
                throw new AppException("アトラクションパスワード(HG)の取得中に予期せぬ応答が返されました(2) [" + response.ResponseUri.Host.ToString() + "]");
            }
            CurrentStateMessage = "アトラクションパスワード(HG)情報の解析中";
            Dictionary<string, string> paramTable = new Dictionary<string, string>();
            using(StreamReader reader = new StreamReader(response.GetResponseStream(), Encoding.GetEncoding(response.CharacterSet))) {
                string line = null;
                Regex startRegex = new Regex("^.*<object id=\"LoadPrgAx\" codebase=\"/nhn/LoadPrgAx.CAB.*\" classid=\"CLSID:.+\" width=.*$");
                Regex endRegex = new Regex("^.*</object>.*$");
                Regex paramRegex = new Regex("^.*<param name=\"(.+)\" value=\"(.+)\" />.*$");
                Match match = null;
                while((line = reader.ReadLine()) != null) {
                    if(startRegex.IsMatch(line)) {
                        // <object>開始から終了までのループ
                        while((line = reader.ReadLine()) != null) {
                            Debug.WriteLine("GetHanGameOneTimePassword#" + line);
                            match = paramRegex.Match(line);
                            if(match.Success) {
                                paramTable.Add(match.Groups[1].Value, match.Groups[2].Value);
                            }
                            if(endRegex.IsMatch(line)) {
                                break;
                            }
                        }
                    }
                }
            }
            response.Close();
            if(!paramTable.ContainsKey("AttractionID") || !paramTable.ContainsKey("Password") || !paramTable.ContainsKey("GameCode")) {
                throw new AppException("アトラクションパスワード(HG)の取得に失敗しました");
            }
            if(!paramTable["GameCode"].StartsWith("c39aa154-55c1-4e0e-92a9-232c3cd5ac61")) {
                throw new AppException("指定されたアトラクションID(HG)はRagnarokOnlineのものではありません");
            }
            targetData.Id = paramTable["AttractionID"];
            targetData.Password = paramTable["Password"];
            CurrentStateMessage = "アトラクションパスワード(HG)の取得が完了";
            return;
        }


        /// <summary>
        /// 初期設定を行ったHttpWebRequestオブジェクトを生成する
        /// </summary>
        /// <param name="url"></param>
        /// <returns></returns>
        private HttpWebRequest CreateHttpWebRequest(string url) {
            HttpWebRequest request = (HttpWebRequest)WebRequest.Create(url);
            request.AllowAutoRedirect = false;
            request.CookieContainer = cookieContainer;
            request.Accept = "*/*";
            request.Method = "GET";
            request.KeepAlive = true;
            request.UserAgent = "Mozilla/5.0 (Windows NT 6.3; WOW64; Trident/7.0; rv:11.0) like Gecko";
            request.ProtocolVersion = new Version(1, 1);
            request.Headers.Add("Accept-Language", "ja-JP");
            request.Headers.Add("UA-CPU", "x86");
            return request;
        }

    }

}
